@@ -40,6 +40,16 @@ CXX_WITH_STDLIB = '''
         }
       '''
 
+CXX_WITH_DYN_CAST = '''
+        struct X { int x; virtual void a() {} };
+        struct Y : X { int y; virtual void a() { y = 10; }};
+        int main(int argc, char **argv) {
+          Y* y = dynamic_cast<Y*>((X*)argv[1]);
+          y->a();
+          return y->y;
+        }
+'''
+
 SYSTEM_TASKS = [
     'al',
     'compiler-rt',
@@ -49,11 +59,15 @@ SYSTEM_TASKS = [
     'libc',
     'libc++',
     'libc++_noexcept',
+    'libc++-mt',
+    'libc++-mt_noexcept',
     'libc++abi',
+    'libc++abi-mt',
     'libc-extras',
     'libc-mt',
     'pthreads',
     'pthreads_stub',
+    'ubsan-rt-wasm',
     'ubsan-minimal-rt-wasm',
 ]
 
@@ -205,7 +219,10 @@ def main():
     auto_tasks = True
   if auto_tasks:
     if shared.Settings.WASM_BACKEND:
-      skip_tasks = [task for task in SYSTEM_TASKS + USER_TASKS if '-mt' in task or 'thread' in task]
+      skip_tasks = []
+      if shared.Settings.RELOCATABLE:
+        # we don't support PIC + pthreads yet
+        skip_tasks += [task for task in SYSTEM_TASKS + USER_TASKS if '-mt' in task or 'thread' in task]
       # cocos2d: must be ported, errors on
       # "Cannot recognize the target platform; are you targeting an unsupported platform?"
       skip_tasks += ['cocos2d']
@@ -263,16 +280,14 @@ def main():
       build(CXX_WITH_STDLIB, ['libc++.a'], ['-s', 'DISABLE_EXCEPTION_CATCHING=0'])
     elif what == 'libc++_noexcept':
       build(CXX_WITH_STDLIB, ['libc++_noexcept.a'])
+    elif what == 'libc++-mt':
+      build(CXX_WITH_STDLIB, ['libc++.a'], ['-s', 'DISABLE_EXCEPTION_CATCHING=0', '-s', 'USE_PTHREADS=1'])
+    elif what == 'libc++-mt_noexcept':
+      build(CXX_WITH_STDLIB, ['libc++_noexcept.a'], ['-s', 'USE_PTHREADS=1'])
     elif what == 'libc++abi':
-      build('''
-        struct X { int x; virtual void a() {} };
-        struct Y : X { int y; virtual void a() { y = 10; }};
-        int main(int argc, char **argv) {
-          Y* y = dynamic_cast<Y*>((X*)argv[1]);
-          y->a();
-          return y->y;
-        }
-      ''', [libname('libc++abi')])
+      build(CXX_WITH_DYN_CAST, [libname('libc++abi')])
+    elif what == 'libc++abi-mt':
+      build(CXX_WITH_DYN_CAST, [libname('libc++abi')], ['-s', 'USE_PTHREADS=1'])
     elif what == 'gl' or what.startswith('gl-'):
       opts = []
       if '-mt' in what:
@@ -313,6 +328,10 @@ def main():
         }
 
       ''', [libname('libpthreads_stub')])
+    elif what == 'ubsan-rt-wasm':
+      if not shared.Settings.WASM_BACKEND:
+        continue
+      build(C_BARE, ['libsanitizer_common_rt_wasm.a', 'libubsan_rt_wasm.a'], ['-fsanitize=undefined', '-s', 'WASM=1'])
     elif what == 'ubsan-minimal-rt-wasm':
       if not shared.Settings.WASM_BACKEND:
         continue
