@@ -23,44 +23,59 @@ function processMacros(text) {
 // ident checked is true in our global.
 // Also handles #include x.js (similar to C #include <file>)
 // Param filenameHint can be passed as a description to identify the file that is being processed, used
-// to locate errors for reporting.
+// to locate errors for reporting and for html files to stop expansion between <style> and </style>.
 function preprocess(text, filenameHint) {
+  var fileExt = (filenameHint) ? filenameHint.split('.').pop().toLowerCase() : "";
+  var isHtml = (fileExt === 'html' || fileExt === 'htm') ? true : false;
+  var inStyle = false;
   var lines = text.split('\n');
   var ret = '';
   var showStack = [];
   for (var i = 0; i < lines.length; i++) {
     var line = lines[i];
     try {
-      if (line[line.length-1] == '\r') {
+      if (line[line.length-1] === '\r') {
         line = line.substr(0, line.length-1); // Windows will have '\r' left over from splitting over '\r\n'
       }
-      if (!line[0] || line[0] != '#') {
-        if (showStack.indexOf(false) == -1) {
-          ret += line + '\n';
-        }
-      } else {
-        if (line[1] == 'i') {
-          if (line[2] == 'f') { // if
-            var parts = line.split(' ');
-            var after = parts.slice(1).join(' ');
-            var truthy = !!eval(after);
-            showStack.push(truthy);
-          } else if (line[2] == 'n') { // include
+      if (isHtml && line.indexOf('<style') !== -1 && !inStyle) {
+        inStyle = true;
+      }
+      if (isHtml && line.indexOf('</style') !== -1 && inStyle) {
+        inStyle = false;
+      }
+
+      if (!inStyle) {
+        if (line.indexOf('#if') === 0) {
+          var parts = line.split(' ');
+          var after = parts.slice(1).join(' ');
+          var truthy = !!eval(after);
+          showStack.push(truthy);
+        } else if (line.indexOf('#include') === 0) {
+          if (showStack.indexOf(false) === -1) {
             var filename = line.substr(line.indexOf(' ')+1);
             if (filename.indexOf('"') === 0) {
               filename = filename.substr(1, filename.length - 2);
             }
             var included = read(filename);
-            ret += '\n' + preprocess(included, filename) + '\n'
+            ret += '\n' + preprocess(included, filename) + '\n';
           }
-        } else if (line[2] == 'l') { // else
-          assert(showStack.length > 0, 'preprocessing error parsing #else on line ' + i + ': ' + line);
+        } else if (line.indexOf('#else') === 0) {
+          assert(showStack.length > 0);
           showStack.push(!showStack.pop());
-        } else if (line[2] == 'n') { // endif
-          assert(showStack.length > 0, 'preprocessing error parsing #endif on line ' + i + ': ' + line);
+        } else if (line.indexOf('#endif') === 0) {
+          assert(showStack.length > 0);
           showStack.pop();
         } else {
-          throw "Unclear preprocessor command on line " + i + ': ' + line;
+          if (line[0] === '#') {
+            throw "Unclear preprocessor command on line " + i + ': ' + line;
+          }
+          if (showStack.indexOf(false) === -1) {
+            ret += line + '\n';
+          }
+        }
+      } else { // !inStyle
+        if (showStack.indexOf(false) === -1) {
+          ret += line + '\n';
         }
       }
     } catch(e) {
@@ -442,7 +457,7 @@ function splitI64(value, floatConversion) {
     asmCoercion('Math_abs(VALUE)', 'double') + ' >= ' + asmEnsureFloat('1', 'double') + ' ? ' +
       '(VALUE > ' + asmEnsureFloat('0', 'double') + ' ? ' +
                asmCoercion('Math_min(' + asmCoercion('Math_floor((VALUE)/' + asmEnsureFloat(4294967296, 'double') + ')', 'double') + ', ' + asmEnsureFloat(4294967295, 'double') + ')', 'i32') + '>>>0' +
-               ' : ' + asmFloatToInt(asmCoercion('Math_ceil((VALUE - +((' + asmFloatToInt('VALUE') + ')>>>0))/' + asmEnsureFloat(4294967296, 'double') + ')', 'double')) + '>>>0' + 
+               ' : ' + asmFloatToInt(asmCoercion('Math_ceil((VALUE - +((' + asmFloatToInt('VALUE') + ')>>>0))/' + asmEnsureFloat(4294967296, 'double') + ')', 'double')) + '>>>0' +
       ')' +
     ' : 0',
     value,
@@ -813,9 +828,6 @@ var asmPrintCounter = 0;
 
 // See makeSetValue
 function makeGetValue(ptr, pos, type, noNeedFirst, unsigned, ignore, align, noSafe, forceAsm) {
-  if (UNALIGNED_MEMORY) align = 1;
-  else if (FORCE_ALIGNED_MEMORY && !isIllegalType(type)) align = 8;
-
   if (isStructType(type)) {
     var typeData = Types.types[type];
     var ret = [];
@@ -894,9 +906,6 @@ function makeGetValueAsm(ptr, pos, type, unsigned) {
 //!             which means we should write to all slabs, ignore type differences if any on reads, etc.
 //! @param noNeedFirst Whether to ignore the offset in the pointer itself.
 function makeSetValue(ptr, pos, value, type, noNeedFirst, ignore, align, noSafe, sep, forcedAlign, forceAsm) {
-  if (UNALIGNED_MEMORY && !forcedAlign) align = 1;
-  else if (FORCE_ALIGNED_MEMORY && !isIllegalType(type)) align = 8;
-
   sep = sep || ';';
   if (isStructType(type)) {
     var typeData = Types.types[type];
@@ -987,7 +996,7 @@ function makeSetValues(ptr, pos, value, type, num, align) {
   if (value < 0) value += 256; // make it unsigned
   var values = {
     1: value,
-    2: value | (value << 8), 
+    2: value | (value << 8),
     4: value | (value << 8) | (value << 16) | (value << 24)
   };
   var ret = [];
@@ -1040,14 +1049,20 @@ function makeHEAPView(which, start, end) {
   return 'HEAP' + which + '.subarray((' + start + ')' + mod + ',(' + end + ')' + mod + ')';
 }
 
-function makeDynCall(sig) {
+// When dynamically linking, some things like dynCalls may not exist in one module and
+// be provided by a linked module, so they must be accessed indirectly using Module
+function exportedAsmFunc(func) {
   if (!MAIN_MODULE && !SIDE_MODULE) {
-    return 'dynCall_' + sig;
+    return func;
   } else {
-    return "Module['dynCall_" + sig + "']";
+    return "Module['" + func + "']";
   }
 }
 
+function makeDynCall(sig) {
+  return exportedAsmFunc('dynCall_' + sig);
+}
+
 var TWO_TWENTY = Math.pow(2, 20);
 
 // Given two values and an operation, returns the result of that operation.
@@ -1442,10 +1457,13 @@ function asmFFICoercion(value, type) {
 }
 
 function makeDynCall(sig) {
-  if (!EMULATED_FUNCTION_POINTERS) {
-    return 'dynCall_' + sig;
-  } else {
+  // asm.js function tables have one table in each linked asm.js module, so we
+  // can't just dynCall into them - ftCall exists for that purpose. In wasm,
+  // even linked modules share the table, so it's all fine.
+  if (EMULATED_FUNCTION_POINTERS && !WASM) {
     return 'ftCall_' + sig;
+  } else {
+    return 'dynCall_' + sig;
   }
 }
 
@@ -1481,6 +1499,26 @@ function makeStaticString(string) {
   return '(stringToUTF8("' + string + '", ' + ptr + ', ' + len + '), ' + ptr + ')';
 }
 
+var ATINITS = [];
+
+function addAtInit(code) {
+  ATINITS.push(code);
+}
+
+var ATMAINS = [];
+
+function addAtMain(code) {
+  ATMAINS.push(code);
+}
+
+var ATEXITS = [];
+
+function addAtExit(code) {
+  if (EXIT_RUNTIME) {
+    ATEXITS.push(code);
+  }
+}
+
 // Generates access to module exports variable in pthreads worker.js. Depending on whether main code is built with MODULARIZE
 // or not, asm module exports need to either be accessed via a local exports object obtained from instantiating the module (in src/worker.js), or via
 // the global Module exports object.
@@ -1521,6 +1559,13 @@ function getQuoted(str) {
 
 function makeRetainedCompilerSettings() {
   var blacklist = set('STRUCT_INFO');
+  if (STRICT) {
+    for (var i in LEGACY_SETTINGS) {
+      var name = LEGACY_SETTINGS[i][0];
+      blacklist[name] = 0;
+    }
+  }
+
   var ret = {};
   for (var x in this) {
     try {
@@ -1539,3 +1584,18 @@ function getPageSize() {
   return WASM ? WASM_PAGE_SIZE : ASMJS_PAGE_SIZE;
 }
 
+// Receives a function as text, and a function that constructs a modified
+// function, to which we pass the parsed-out name, arguments, and body of the
+// function. Returns the output of that function.
+function modifyFunction(text, func) {
+  var match = text.match(/\s*function\s+([^(]*)?\s*\(([^)]*)\)/);
+  assert(match, 'could not match function ' + text + '.');
+  var name = match[1];
+  var args = match[2];
+  var rest = text.substr(match[0].length);
+  var bodyStart = rest.indexOf('{');
+  assert(bodyStart > 0);
+  var bodyEnd = rest.lastIndexOf('}');
+  assert(bodyEnd > 0);
+  return func(name, args, rest.substring(bodyStart + 1, bodyEnd));
+}
